import numpy as np
from onnxruntime import InferenceSession
from transformers import AutoTokenizer
import re

class OmographModel:
    def __init__(self):
        pass

    def load(self, path, device="CPU"):
        self.session = InferenceSession(f"{path}/model.onnx", providers=["CUDAExecutionProvider" if device == "CUDA" else "CPUExecutionProvider"])
        self.tokenizer = AutoTokenizer.from_pretrained(path)
        
    def softmax(self, x):
        e_x = np.exp(x - np.max(x))
        return e_x / e_x.sum()

    def group_words(self, words):
        groups = {}
        for word in words:
            parts = word.replace('+', '')
            key = parts
            groups.setdefault(key, []).append(word)
        return list(groups.values())
        
    def transfer_grouping(self, grouped_list, target_list):
        new_grouped_list = []
        start_index = 0
        for group in grouped_list:
            group_length = len(group)
            new_group = target_list[start_index:start_index + group_length]
            new_grouped_list.append(new_group)
            start_index += group_length
        return new_grouped_list
        
    def classify(self, texts, hypotheses):
        hypotheses_probs = []
        preprocessed_texts = [re.sub(r'\s+(?=(?:[,.?!:;…]))', r'', text) for text in texts]
        if len(hypotheses) % 2 != 0:
            outs = []
            grouped_h = self.group_words(hypotheses)
            grouped_t = self.transfer_grouping(grouped_h, preprocessed_texts)
            for h, t in zip(grouped_h, grouped_t):
                probs = []
                for hp in h:
                    inputs = self.tokenizer(t[0], hp, max_length=512, truncation=True, return_tensors="np")
                    inputs = {k: v.astype(np.int64) for k, v in inputs.items()}
                    outputs = self.session.run(None, inputs)[0]
                    outputs = self.softmax(outputs)
                    prob_label_is_true = [float(p[1]) for p in outputs][0]
                    probs.append(prob_label_is_true)
                    #print(h, prob_label_is_true)
                outs.append(h[probs.index(max(probs))])
            return outs
        else:
            inputs = self.tokenizer(preprocessed_texts, hypotheses, return_tensors="np", padding=True, truncation=True, max_length=512)
            inputs = {k: v.astype(np.int64) for k, v in inputs.items()}
    
            outputs = self.session.run(None, inputs)[0]
            outputs = self.softmax(outputs)
            #print(hypotheses)
            preprocessed_texts = [(preprocessed_texts[i], preprocessed_texts[i+1]) for i in range(0, len(preprocessed_texts), 2)]
            hypotheses =  [(hypotheses[i], hypotheses[i+1]) for i in range(0, len(hypotheses), 2)]
            
            for i in range(len(texts)):
                prob_label_is_true = float(outputs[i][1])
                hypotheses_probs.append(prob_label_is_true)
    
            hypotheses_probs = [(hypotheses_probs[i], hypotheses_probs[i+1]) for i in range(0, len(hypotheses_probs), 2)]
            outs = []
            for pair1, pair2 in zip(hypotheses, hypotheses_probs):
              outs.append(pair1[pair2.index(max(pair2))])
            return outs


import json
import pathlib
from huggingface_hub import HfFileSystem, hf_hub_download
import os
import gzip
from os.path import join as join_path
from .omograph_model import OmographModel
from .accent_model import AccentModel
from .stress_usage_model import StressUsagePredictorModel
from .yo_homograph_model import YoHomographModel
from .text_split import split_by_sentences
from .rule_accent_engine import RuleEngine
import re
from .rumorph import RUMorph



class RUAccent:
    def __init__(self):
        self.omograph_model = OmographModel()
        self.accent_model = AccentModel()
        self.stress_usage_predictor = StressUsagePredictorModel()
        self.yo_homograph_model = YoHomographModel()
        self.fs = HfFileSystem()
        self.rule_accent = RuleEngine()
        self.rumorph = RUMorph()
        self.omograph_models_paths = {'big': '/nn/nn_omograph/big', 
                                      'medium': '/nn/nn_omograph/medium', 
                                      'small': '/nn/nn_omograph/small', 
                                      'big_poetry': '/nn/nn_omograph/big_poetry', 
                                      'medium_poetry': '/nn/nn_omograph/medium_poetry', 
                                      'small_poetry': '/nn/nn_omograph/small_poetry'}
    
        self.accentuator_paths = ['/nn/nn_accent', '/nn/nn_stress_usage_predictor','/nn/nn_yo_homograph_resolver', '/dictionary', '/dictionary/rule_engine', '/dictionary/lemmatizer', '/nn/nn_morph/morph', '/nn/nn_morph/syntax']
        self.letters_accent = {'о': '+о', 'О': '+О'}

    def load(
        self,
        omograph_model_size="big_poetry",
        use_dictionary=False,
        custom_dict={},
        custom_homographs={},
        repo="ruaccent/accentuator",
        workdir=None
        ):
        if workdir:
            self.workdir = workdir
        else:
            self.workdir = str(pathlib.Path(__file__).resolve().parent)

        self.custom_dict = custom_dict
        self.accents = {}
        if not os.path.exists(
            join_path(self.workdir, "dictionary")
        ):
            for path in self.accentuator_paths:
                files = self.fs.ls(repo + path)
                print(files)
                for file in files:
                    if file["type"] == "file":
                        hf_hub_download(repo_id=repo, local_dir_use_symlinks=False, local_dir=self.workdir, filename=file['name'].replace(repo+'/', ''))
    
        if not os.path.exists(join_path(self.workdir, "nn")):
            os.mkdir(join_path(self.workdir, "nn"))
        
        if not os.path.exists(join_path(self.workdir, "nn", "nn_omograph", omograph_model_size)):
            model_path = self.omograph_models_paths.get(omograph_model_size, None)
            if model_path:
                files = self.fs.ls(repo + model_path)
                for file in files:
                    hf_hub_download(repo_id=repo, local_dir_use_symlinks=False, local_dir=self.workdir, filename=file['name'].replace(repo+'/', ''))
            else:
                raise FileNotFoundError
        
        self.omographs = json.load(
            gzip.open(join_path(self.workdir, "dictionary","omographs.json.gz"))
        )
        self.omographs.update(custom_homographs)
        self.yo_words = json.load(
            gzip.open(join_path(self.workdir, "dictionary","yo_words.json.gz"))
        )

        if use_dictionary:
            self.accents.update(json.load(
                gzip.open(join_path(self.workdir, "dictionary","accents.json.gz"))
            ))

        self.accents.update(self.custom_dict)
        self.accents.update(self.letters_accent)
        self.omograph_model.load(
            join_path(self.workdir, f"nn/nn_omograph/{omograph_model_size}/")
        )
        self.yo_homographs = json.load(
            gzip.open(join_path(self.workdir, "dictionary","yo_homographs.json.gz"))
        )
        self.accent_model.load(join_path(self.workdir, "nn","nn_accent/"))
        self.stress_usage_predictor.load(join_path(self.workdir, "nn","nn_stress_usage_predictor/"))
        self.yo_homograph_model.load(join_path(self.workdir, "nn","nn_yo_homograph_resolver"))
        self.rumorph.load(path=join_path(self.workdir, "nn","nn_morph"), lemmatizer_path=join_path(self.workdir, "dictionary","lemmatizer"))
        self.rule_accent.load(join_path(self.workdir, "dictionary","rule_engine"), self.rumorph)

    def split_by_words(self, string):
        string = string.replace(" - ",' ~ ')
        result = re.findall(r"\w*(?:\+\w+)*|[^\w\s]+", string.lower())
        return [res for res in result if res]


    def count_vowels(self, text):
        vowels = "аеёиоуыэюяАЕЁИОУЫЭЮЯ"
        return sum(1 for char in text if char in vowels)

    def has_punctuation(self, text):
        for char in text:
            if char in "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~":
                return True
        return False

    def delete_spaces_before_punc(self, text):
        punc = "!\"#$%&'()*,./:;<=>?@[\\]^_`{|}-"
        for char in punc:
            if char == '-':
                text = text.replace(" " + char, char).replace(char + " ", char)
            text = text.replace(" " + char, char)
        return text.replace('~', '-')

    def extract_entities(self, data):
        entities = []
        for item in data:
            entity = item['entity']
            entities.append(entity)
        return entities

    def _process_yo(self, text, sentence):
        splitted_text = text
        yo_predictions = self.extract_entities(self.yo_homograph_model.predict_yo_homographs(sentence))
        
        for i, word in enumerate(splitted_text):
            splitted_text[i] = self.yo_words.get(word, word)
            if yo_predictions[i] == "YO":
                splitted_text[i] = self.yo_homographs.get(word, word)
        return splitted_text

    def _process_omographs(self, text, sentence):
        splitted_text = text
    
        founded_omographs = []
        texts = []
        hypotheses = []
    
        for i, word in enumerate(splitted_text):
            variants = self.omographs.get(word)
            if variants:
                founded_omographs.append(
                    {"word": word, "variants": variants, "position": i}
                )
                texts.append(splitted_text)
                hypotheses.append(variants)
    
        if len(founded_omographs) > 0:
            texts_batch = []
            hypotheses_batch = [val for sublist in hypotheses for val in sublist]

            for o, t in zip(founded_omographs, texts):
                position = o["position"]
                t_back = t[position]
                t[position] = ' <w>' + t[position] + '</w> '
                texts_batch.append(self.delete_spaces_before_punc(" ".join(t.copy())))
                texts_batch.append(self.delete_spaces_before_punc(" ".join(t.copy())))
                t[position] = t_back
            cls_batch = self.omograph_model.classify(texts_batch, hypotheses_batch)
    
            cls_index = 0
            for omograph in founded_omographs:
                position = omograph["position"]
                splitted_text[position] = cls_batch[cls_index]
                cls_index += 1
    
        return splitted_text


    
    def _process_accent(self, text, stress_usages):
        splitted_text = text

        for i, word in enumerate(splitted_text):
            if stress_usages[i] == "STRESS":
                stressed_word = self.accents.get(word, word)
                if stressed_word == word and not self.has_punctuation(word) and self.count_vowels(word) > 1:
                    splitted_text[i] = self.accent_model.put_accent(word)
                else:
                    splitted_text[i] = stressed_word
        return splitted_text

        
    def process_yo(self, text):
        sentences = split_by_sentences(text)
        outputs = []
        for sentence in sentences:
            text = self.split_by_words(sentence)
            processed_text = self._process_yo(text)
            processed_text = " ".join(processed_text)
            processed_text = self.delete_spaces_before_punc(processed_text)
            outputs.append(processed_text)
        return " ".join(outputs)
    
    def process_all(self, text):
        sentences = split_by_sentences(text)
        outputs = []
        for sentence in sentences:
            text = self.split_by_words(sentence)
            accented_tokens = self.rule_accent.accentuate(sentence)
            if len(accented_tokens) == len(text):
                for i in range(len(text)):
                    if '+' in accented_tokens[i]:
                        text[i] = accented_tokens[i]
            stress_usages = self.extract_entities(self.stress_usage_predictor.predict_stress_usage(sentence))
            processed_text = self._process_yo(text, sentence)
            processed_text = self._process_omographs(processed_text, sentence)
            processed_text = self._process_accent(processed_text, stress_usages)
            processed_text = " ".join(processed_text)
            processed_text = self.delete_spaces_before_punc(processed_text)
            outputs.append(processed_text)
        return " ".join(outputs)



# -*- coding: utf-8 -*-
import json
import gzip
DEBUG = False

def decode_pos(pos):
    if pos in ['ДЕЕПРИЧАСТИЕ', 'ГЛАГОЛ', 'ИНФИНИТИВ']:
        return 'ГЛАГОЛ'
    else:
        return pos


class Lemmatizer(object):
    def __init__(self):
        pass

    def load(self, dict_path):
        self.forms = json.load(gzip.open(dict_path+"/forms.json.gz","r"))
        self.forms2 = json.load(gzip.open(dict_path+"/forms2.json.gz","r"))
        self.special_lemmas = json.load(gzip.open(dict_path+"/spec.json.gz","r"))
        self.key2transducer = json.load(gzip.open(dict_path+"/key2transducer.json.gz","r"))

    def get_lemma(self, word):
        if word in self.forms:
            return self.forms[word]
        elif word in self.forms2:
            return self.forms2[word][0]
        elif word in self.special_lemmas:
            return self.special_lemmas[word]
        else:
            return word

    def decode_pos_tags(self, pos_tags):
        stags1 = []
        part_of_speech = 'unk'
        short_tag_index = -1
        for tag in pos_tags.split('|'):
            if tag == 'NOUN':
                part_of_speech = 'СУЩЕСТВИТЕЛЬНОЕ'
            elif tag == 'VERB':
                part_of_speech = 'ГЛАГОЛ'
            elif tag == 'ADJ':
                part_of_speech = 'ПРИЛАГАТЕЛЬНОЕ'
                stags1.append(('КРАТКИЙ', '0'))
                short_tag_index = 0
            elif tag == 'ADV':
                part_of_speech = 'НАРЕЧИЕ'
            elif tag == 'PRON':
                part_of_speech = 'МЕСТОИМЕНИЕ'
            elif tag == 'ADP':
                part_of_speech = 'ПРЕДЛОГ'
            elif tag == 'CONJ':
                part_of_speech = 'СОЮЗ'
            elif tag == 'PART':
                part_of_speech = 'ЧАСТИЦА'
            elif tag == 'PUNCT':
                part_of_speech = 'ПУНКТУАТОР'
            elif '=' in tag:
                if part_of_speech == 'СУЩЕСТВИТЕЛЬНОЕ':
                    if tag == 'Case=Nom':
                        stags1.append(('ПАДЕЖ', 'ИМ'))
                    elif tag == 'Case=Acc':
                        stags1.append(('ПАДЕЖ', 'ВИН'))
                    elif tag == 'Case=Dat':
                        stags1.append(('ПАДЕЖ', 'ДАТ'))
                    elif tag == 'Case=Ins':
                        stags1.append(('ПАДЕЖ', 'ТВОР'))
                    elif tag == 'Case=Prep':
                        stags1.append(('ПАДЕЖ', 'ПРЕДЛ'))
                    elif tag == 'Case=Loc':
                        stags1.append(('ПАДЕЖ', 'ПРЕДЛ'))  # 03-02-2020 'МЕСТ'
                    elif tag == 'Case=Gen':
                        stags1.append(('ПАДЕЖ', 'РОД'))
                    elif tag == 'Case=Voc':
                        stags1.append(('ПАДЕЖ', 'ЗВАТ'))
                    elif tag == 'Number=Sing':
                        stags1.append(('ЧИСЛО', 'ЕД'))
                    elif tag == 'Number=Plur':
                        stags1.append(('ЧИСЛО', 'МН'))
                    elif tag == 'Gender=Masc':
                        stags1.append(('РОД', 'МУЖ'))
                    elif tag == 'Gender=Fem':
                        stags1.append(('РОД', 'ЖЕН'))
                    elif tag == 'Gender=Neut':
                        stags1.append(('РОД', 'СР'))
                    else:
                        if DEBUG:
                            print('неизвестный тэг "{}"'.format(tag))
                        #raise NotImplementedError()
                elif part_of_speech == 'ПРИЛАГАТЕЛЬНОЕ':
                    if tag == 'Case=Nom':
                        stags1.append(('ПАДЕЖ', 'ИМ'))
                    elif tag == 'Case=Acc':
                        stags1.append(('ПАДЕЖ', 'ВИН'))
                    elif tag == 'Case=Dat':
                        stags1.append(('ПАДЕЖ', 'ДАТ'))
                    elif tag == 'Case=Ins':
                        stags1.append(('ПАДЕЖ', 'ТВОР'))
                    elif tag == 'Case=Prep':
                        stags1.append(('ПАДЕЖ', 'ПРЕДЛ'))
                    elif tag == 'Case=Loc':
                        stags1.append(('ПАДЕЖ', 'ПРЕДЛ')) # 03-02-2020 'МЕСТ'
                    elif tag == 'Case=Gen':
                        stags1.append(('ПАДЕЖ', 'РОД'))
                    elif tag == 'Number=Sing':
                        stags1.append(('ЧИСЛО', 'ЕД'))
                    elif tag == 'Number=Plur':
                        stags1.append(('ЧИСЛО', 'МН'))
                    elif tag == 'Gender=Masc':
                        stags1.append(('РОД', 'МУЖ'))
                    elif tag == 'Gender=Fem':
                        stags1.append(('РОД', 'ЖЕН'))
                    elif tag == 'Gender=Neut':
                        stags1.append(('РОД', 'СР'))
                    elif tag == 'Degree=Cmp':
                        stags1.append(('СТЕПЕНЬ', 'СРАВН'))
                    elif tag == 'Degree=Pos':
                        stags1.append(('СТЕПЕНЬ', 'АТРИБ'))
                    elif tag in ('Variant=Short', 'Variant=Brev'):
                        stags1[short_tag_index] = ('КРАТКИЙ', '1')
                    else:
                        if DEBUG:
                            print('неизвестный тэг "{}"'.format(tag))
                            #raise NotImplementedError()
                elif part_of_speech == 'ГЛАГОЛ':
                    #print("#"+tag+"#")
                    if tag == 'Number=Sing':
                        stags1.append(('ЧИСЛО', 'ЕД'))
                    elif tag == 'Number=Plur':
                        stags1.append(('ЧИСЛО', 'МН'))
                    elif tag == 'Gender=Masc':
                        stags1.append(('РОД', 'МУЖ'))
                    elif tag == 'Gender=Fem':
                        stags1.append(('РОД', 'ЖЕН'))
                    elif tag == 'Gender=Neut':
                        stags1.append(('РОД', 'СР'))
                    elif tag == 'Mood=Ind':
                        stags1.append(('НАКЛОНЕНИЕ', 'ИЗЪЯВ'))
                    elif tag == 'Mood=Imp':
                        stags1.append(('НАКЛОНЕНИЕ', 'ПОБУД'))
                    elif tag == 'Tense=Past':
                        stags1.append(('ВРЕМЯ', 'ПРОШЕДШЕЕ'))
                    elif tag == 'Tense=Fut':
                        stags1.append(('ВРЕМЯ', 'БУДУЩЕЕ'))
                    elif tag == 'Tense=Notpast':
                        stags1.append(('ВРЕМЯ', 'НАСТОЯЩЕЕ'))
                    elif tag == 'Tense=Pres':
                        stags1.append(('ВРЕМЯ', 'НАСТОЯЩЕЕ'))
                    elif tag == 'Person=1':
                        stags1.append(('ЛИЦО', '1'))
                    elif tag == 'Person=2':
                        stags1.append(('ЛИЦО', '2'))
                    elif tag == 'Person=3':
                        stags1.append(('ЛИЦО', '3'))
                    elif tag == 'VerbForm=Fin':
                        pass
                    elif tag == 'VerbForm=Inf':
                        pass
                    elif tag == 'VerbForm=Conv':
                        pass
                    else:
                        if DEBUG:
                            print('неизвестный тэг "{}"'.format(tag))
                        #raise RuntimeError(msg)
                elif part_of_speech == 'НАРЕЧИЕ':
                    if tag == 'Degree=Pos':
                        stags1.append(('СТЕПЕНЬ', 'АТРИБ'))
                    elif tag == 'Degree=Cmp':
                        stags1.append(('СТЕПЕНЬ', 'СРАВН'))
                    else:
                        if DEBUG:
                            print('неизвестный тэг "{}"'.format(tag))
                    pass

        return part_of_speech, stags1

    def get_lemma2(self, word, pos_tags):
        part_of_speech, decoded_tags = self.decode_pos_tags(pos_tags)

        nword = word.lower().replace('ё', 'е')

        if nword in self.special_lemmas:
            return self.special_lemmas[nword], part_of_speech, decoded_tags

        if nword in self.forms:
            lemma = self.forms[nword]
            return lemma, part_of_speech, decoded_tags
        elif nword in self.forms2:
            if part_of_speech == 'СУЩЕСТВИТЕЛЬНОЕ':
                # Для существительных учитываем падеж.
                required_case = None
                for tag in decoded_tags:
                    if tag[0] == 'ПАДЕЖ':
                        required_case = tag[1]
                        break

                for lemma, lemma_part_of_speech, tag in self.forms2[nword]:
                    if lemma_part_of_speech == part_of_speech and tag == required_case:
                        return lemma, part_of_speech, decoded_tags
            else:
                for lemma, lemma_part_of_speech, tags in self.forms2[nword]:
                    if lemma_part_of_speech == part_of_speech:
                        return lemma, part_of_speech, decoded_tags
        elif len(word) > 4:
            ending = nword[-4:]
            key = ending + '|' + part_of_speech
            if key in self.key2transducer:
                transducer = self.key2transducer[key]
                if transducer[0] > 0:
                    lemma = word[:-transducer[0]] + transducer[1]
                else:
                    lemma = word + transducer[1]

                return lemma.lower(), part_of_speech, decoded_tags

        return nword, part_of_speech, decoded_tags

    def lemmatize(self, tagged_words):
        return [(word, tags,)+tuple(self.get_lemma2(word, tags)) for (word, tags) in tagged_words]



from .tags_precitor import TagsPredictor
from .morph_precitor import MorphPredictor
from .lemmatizer import Lemmatizer


morph = {
  "=ADJ": 83832,
  "=ADP": 83843,
  "=ADV": 83842,
  "=CONJ": 83837,
  "=DET": 83836,
  "=H": 83828,
  "=INTJ": 83833,
  "=LATN": 83831,
  "=NOUN": 83840,
  "=NUM": 83839,
  "=PART": 83838,
  "=PRON": 83830,
  "=PUNCT": 83841,
  "=SYM": 83834,
  "=VERB": 83835,
  "=X": 83829
}.keys()

class RUMorph:
    def __init__(self) -> None:
        self.t = TagsPredictor()
        self.m = MorphPredictor()
        self.lemmatizer = Lemmatizer()
    def load(self, path, lemmatizer_path):
        self.t.load(path+"/syntax")
        self.m.load(path+"/morph")
        self.lemmatizer.load(lemmatizer_path)

    def _convert_lists(self, list1, list2):
        converted_list = []
        for item1, item2 in zip(list1, list2):
            converted_item = {
                "word": item1["word"],
                "tags": item2["entity"] + "|" + item1["entity"]
            }
            converted_list.append(converted_item)
        return converted_list
    def _convert_morph(self, list1):
        out = []
        for chunk in list1:
            out.append(chunk["word"].strip() + "=" + chunk["entity"])
        return out
    
    def _convert_out(self, list_1):
        out = []
        for chunk in list_1:
            if chunk["word"] in morph:
                out[-1] = out[-1] + chunk["word"]
            else:
                out.append(chunk["word"])
        for i in range(len(out)):
            out[i] = out[i] + "|" + list_1[i]['entity']
        return [item.split('=',1) for item in out]
    

    def process(self, text):
        processed2 = self._convert_morph(self.m.predict_morph(text))
        processed1 = self.t.predict_tags(processed2)
        lemmas = self.lemmatizer.lemmatize(self._convert_out(processed1)) #self._convert_lists(processed1, processed2)
        return [
            [word, tags, lemma]
            for word, tags, lemma, *_ in lemmas
        ]